/****************************************************************/
/*  
     ***   ***  ***  *** ***     *** ***   ******  ****  ***
     ***   ***  ***  *** ***     *** **   ***  *** ***** ***
     ***   ***  ***  *** ***     *****    ******** *** *****
      *** ***   ***  *** ******* *** **   ***  *** ***  ****
       *****     *****   ******* *** ***  ***  *** ***   ***
	   
	 Vulkan Graphics API: in 20 Minutes (Coffee Break Series)
	 Getting Started

     Minimal working examples (MWE) to build upon and 
	 modify to gain an understanding of the API

	 Extensions to the Vulkan Graphics Coffee Break Series
	   
/*
/****************************************************************/
/*
     About:
	 helper.h set of helper functions to reduce the complexity
	 of the `demos' without introducing a complete framework.

	 Simple math classes, e.g., Vector, Matrix,

	 Simple wrapper classes for initiating Vulkan, loading
	 shaders, etc.  The majority of the code is still inside
	 main.cpp.

     You also need to compile and include the shader
	 files (vertex and pixel shaders).
	 e.g., run the compilershaders.bat

	 Extension Chapters 0X - Multiple Renders

	 Rendering to textures and using them in the main 
	 scene.
*/

#pragma once




#include "vulkan.h"

#pragma comment(linker, "/subsystem:windows")
#include <windows.h>



// Debug defines (custom asserts)
#if defined(_WIN32)
  #define DBG_ASSERT(f) { if(!(f)){ __debugbreak(); }; }
#else
  #define DBG_ASSERT(f) { if(!(f)){ DBG_ASSERT(0); }; }
#endif

#define DBG_VALID(f)  { if( (f)!=(f) ){DBG_ASSERT(false);} }

#define DBG_ASSERT_MSG( val, errmsg )        \
	if(!(val)){                              \
	DBG_ASSERT( false )                      \
	}

#define DBG_ASSERT_VULKAN( val )             \
    DBG_ASSERT( VK_SUCCESS == val )

#define DBG_ASSERT_VULKAN_MSG( val, errmsg ) \
	if(!((VK_SUCCESS == val))){              \
    DBG_ASSERT( false )                      \
	}

// Check return result from Vulkan calls
#define VK_CHECK_RESULT(f)					 \
{											 \
	VkResult res = (f);						 \
	if (res != VK_SUCCESS)					 \
	{										 \
		DBG_ASSERT(false); 					 \
	}										 \
}	







/**************************************************************/
/*
   Minimal vector class of 3 floats and overloaded math operators
*/
/**************************************************************/

class Vector3 
{	
public:
	float x, y, z;

	Vector3	(float x, float y, float z)			{this->x=x; this->y=y; this->z=z;}
	Vector3	()									{}

	void	operator+= (const Vector3 &v)		{x+=v.x;   y+=v.y;   z+=v.z;}
	Vector3 operator/  (const float &v)	  const	{return Vector3(x/v,y/v,z/v);}
	Vector3 operator-  (const Vector3 &v) const	{return Vector3(x-v.x,y-v.y,z-v.z);}
	Vector3 operator+  (const Vector3 &v) const {return Vector3(x+v.x,y+v.y,z+v.z);}
	Vector3 operator*  (const float &v)   const	{return Vector3(x*v,y*v,z*v);}
	Vector3 operator-  ()				  const	{return Vector3(-x,-y,-z);}

	static Vector3  Cross	(const Vector3 &vA, const Vector3& vB)	
	{
		return Vector3(vA.y*vB.z - vA.z*vB.y, vA.z*vB.x - vA.x*vB.z, vA.x*vB.y - vA.y*vB.x);
	}
	static float	Dot		(const Vector3 &vA, const Vector3& vB)	{return vA.x*vB.x + vA.y*vB.y + vA.z*vB.z;}
	static Vector3 Normalize(const Vector3& v)
	{
		const float l = v.Length();
		return Vector3(v.x/l,v.y/l,v.z/l);
	}
	float	Length()	 const					{return sqrt(x*x+y*y+z*z);}
	float	LengthSq()	 const					{return (x*x+y*y+z*z);}

	static  double Length(const Vector3& v)
	{
		return v.Length();
	}
	
}; // End Vector3(..)



/**************************************************************/
/*
   Minimal matrix class basic operators and overloaded math
   functions
*/
/**************************************************************/

class Matrix4
{
public:
	float m[16];

	Matrix4::Matrix4()									
	{
		// Don't do anything in here - as it can be expensive to 
		// keep initializing Matrix4s (i.e., large arrays or reading
		// data from files)
	}


	// Copy constructor
	Matrix4::Matrix4(const Matrix4& rhs)
	{
		for (int i=0; i<16; ++i)
		{
			m[i] = rhs.m[i];
		}// End for i
	}

	Matrix4::Matrix4(	float _00, float _01, float _02, float _03,
						float _10, float _11, float _12, float _13,
						float _20, float _21, float _22, float _23,
						float _30, float _31, float _32, float _33)
	{
		m[0]=_00;	m[4]=_01;	m[8] =_02;	m[12]=_03;
		m[1]=_10;	m[5]=_11;	m[9] =_12;	m[13]=_13;
		m[2]=_20;	m[6]=_21;	m[10]=_22;	m[14]=_23;
		m[3]=_30;   m[7]=_31;	m[11]=_32;	m[15]=_33;
	
		/* 
		// For row-order
	m[0] = _00;	m[4]  = _10;	m[8]  = _20;	m[12] = _30;
	m[1] = _01;	m[5]  = _11;	m[9]  = _21;	m[13] = _31;
	m[2] = _02;	m[6]  = _12;	m[10] = _22;	m[14] = _32;
	m[3] = _03;	m[7]  = _13;	m[11] = _23;	m[15] = _33;
		*/
	}



	float&			Matrix4::operator[]	(int index)			{ return m[index]; }

	const float&	Matrix4::operator[]	(int index) const	{ return m[index]; }


	static Matrix4 Matrix4::Identity()
	{
		static Matrix4 identity(1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1);
		return identity;
	}

	static 
	Matrix4 Matrix4::SetTranslation(const Vector3& translation)
	{
		Matrix4 mat = Matrix4::Identity();
		mat.m[12] = translation.x;
		mat.m[13] = translation.y;
		mat.m[14] = translation.z;
		return mat;
	}

	static Matrix4 SetScale			(const Vector3& scale)
	{
		Matrix4 mat = Matrix4::Identity();
		mat.m[ 0] = scale.x;
		mat.m[ 5] = scale.y;
		mat.m[10] = scale.z;
		return mat;
	}

	// ref: https://en.wikipedia.org/wiki/Rotation_matrix#Rotation_matrix_from_axis_and_angle
	static Matrix4 Matrix4::SetRotationAxis(float angle, const Vector3& axis)
	{
		// Page 466, Graphics Gems

		float s = sin(angle);
		float c = cos(angle);
		float t = 1 - c;

		Vector3 ax = Vector3::Normalize( axis );

		float x = ax.x;
		float y = ax.y;
		float z = ax.z;

		Matrix4 rotate(t*x*x+c, t*x*y-s*z, t*x*z+s*y, 0,
						t*y*x+s*z, t*y*y+c, t*y*z-s*x, 0,
						t*z*x-s*y, t*z*y+s*x, t*z*z+c, 0,
						0,			0,			0,		1 );

		return rotate;
	}

	// https://msdn.microsoft.com/en-us/library/windows/desktop/bb205342(v=vs.85).aspx
	// Left Handed
	//2*zn/w  0       0              0
	//0       2*zn/h  0              0
	//0       0       zf/(zf-zn)     1
	//0       0       zn*zf/(zn-zf)  0

	// https://msdn.microsoft.com/en-us/library/windows/desktop/bb205355(v=vs.85).aspx
	// Right Handed
	//2*zn/w  0       0              0
	//0       2*zn/h  0              0
	//0       0       zf/(zn-zf)    -1
	//0       0       zn*zf/(zn-zf)  0

	static 
	Matrix4 SetPerspectiveRH(float FOV, float Aspect, float ZNear, float ZFar)
	{
		float Width =    1.0f / tanf(FOV/2.0f);
		float Height = Aspect / tanf(FOV/2.0f);

		Matrix4 res;
		res[ 0] = Width;
		res[ 4] = 0.0f;
		res[ 8] = 0.0f;
		res[12] = 0.0f;

		res[ 1] = 0.0f;
		res[ 5] = -Height;
		res[ 9] = 0.0f;
		res[13] = 0.0f;
	
		res[ 2] = 0.0f;
		res[ 6] = 0.0f;
		res[10] = ZFar / (ZNear - ZFar);
		res[14] = ZFar * ZNear / (ZNear - ZFar);
	
		res[ 3] = 0.0f;
		res[ 7] = 0.0f;
		res[11] = -1.0f;
		res[15] = 0.0f;
		return res;
	}

	// https://msdn.microsoft.com/en-us/library/windows/desktop/bb205342(v=vs.85).aspx
	// Left Handed
	// zaxis = normal(at - eye)
	// xaxis = normal(cross(Up, zaxis))
	// yaxis = cross(zaxis, xaxis)
	// xaxis.x           yaxis.x           zaxis.x          0
	// xaxis.y           yaxis.y           zaxis.y          0
	// xaxis.z           yaxis.z           zaxis.z          0
	//-dot(xaxis, eye)  -dot(yaxis, eye)  -dot(zaxis, eye)  1

	// https://msdn.microsoft.com/en-us/library/windows/desktop/bb205343(v=vs.85).aspx
	// Right Handed
	//zaxis = normal(eye - at)
	//xaxis = normal(cross(Up, zaxis))
	//yaxis = cross(zaxis, xaxis)
	//xaxis.x           yaxis.x           zaxis.x          0
	//xaxis.y           yaxis.y           zaxis.y          0
	//xaxis.z           yaxis.z           zaxis.z          0
	//dot(xaxis, eye)   dot(yaxis, eye)   dot(zaxis, eye)  1

	static 
	Matrix4 SetLookAtRH(const Vector3 &eye, const Vector3 &at, const Vector3 &up)
	{
		Vector3 XAxis, YAxis, ZAxis;
		ZAxis = Vector3::Normalize(eye - at);
		XAxis = Vector3::Normalize(Vector3::Cross(up, ZAxis));
		YAxis = Vector3::Normalize(Vector3::Cross(ZAxis, XAxis));

		Matrix4 res;
		res[ 0] = XAxis.x;
		res[ 4] = XAxis.y;
		res[ 8] = XAxis.z;
		res[12] = -Vector3::Dot(XAxis,eye);
		
		res[ 1] = YAxis.x;
		res[ 5] = YAxis.y;
		res[ 9] = YAxis.z;
		res[13] = -Vector3::Dot(YAxis,eye);
		
		res[ 2] = ZAxis.x;
		res[ 6] = ZAxis.y;
		res[10] = ZAxis.z;
		res[14] = -Vector3::Dot(ZAxis,eye);
		
		res[ 3] = 0.0f;
		res[ 7] = 0.0f;
		res[11] = 0.0f;
		res[15] = 1.0f;
		return res;
	}

	// https://msdn.microsoft.com/en-us/library/windows/desktop/bb205349(v=vs.85).aspx
	// Right Handed
	//2/w  0    0           0
	//0    2/h  0           0
	//0    0    1/(zn-zf)   0
	//0    0    zn/(zn-zf)  1

	// https://msdn.microsoft.com/en-us/library/windows/desktop/bb205346(v=vs.85).aspx
	// Left Handed
	//2/w  0    0           0
	//0    2/h  0           0
	//0    0    1/(zf-zn)   0
	//0    0   -zn/(zf-zn)  1

	static Matrix4 SetOrthogonal(float Width, float Height, float ZNear, float ZFar)
	{
		float l = 0;
		float r = Width;
		float t = 0;
		float b = Height;

		Matrix4 res;
		res[ 0] = 2.0f / (r-l);
		res[ 1] = 0.0f;
		res[ 2] = 0.0f;
		res[ 3] = 0.0f;

		res[ 4] = 0.0f;
		res[ 5] = 2.0f / (t-b);
		res[ 6] = 0.0f;
		res[ 7] = 0.0f;

		res[ 8] = 0.0f;
		res[ 9] = 0.0f;
		res[10] = -2.0f / (ZFar - ZNear);
		res[11] = 0;

		res[12] = -(r+l)/(r-l);
		res[13] = -(t+b)/(t-b);
		res[14] = -(ZFar+ZNear)/(ZFar-ZNear);
		res[15] = 1.0f;
		return res;
	}
	
}; // End Matrix4(..)




//inline
Matrix4 operator*(const Matrix4& lhs, const Matrix4& rhs)
{
	Matrix4 result;

	result[0] = lhs[0]*rhs[0] + lhs[4]*rhs[1] + lhs[8]*rhs[2] + lhs[12]*rhs[3];
	result[1] = lhs[1]*rhs[0] + lhs[5]*rhs[1] + lhs[9]*rhs[2] + lhs[13]*rhs[3];
	result[2] = lhs[2]*rhs[0] + lhs[6]*rhs[1] + lhs[10]*rhs[2] + lhs[14]*rhs[3];
	result[3] = lhs[3]*rhs[0] + lhs[7]*rhs[1] + lhs[11]*rhs[2] + lhs[15]*rhs[3];
	
	result[4] = lhs[0]*rhs[4] + lhs[4]*rhs[5] + lhs[8]*rhs[6] + lhs[12]*rhs[7];
	result[5] = lhs[1]*rhs[4] + lhs[5]*rhs[5] + lhs[9]*rhs[6] + lhs[13]*rhs[7];
	result[6] = lhs[2]*rhs[4] + lhs[6]*rhs[5] + lhs[10]*rhs[6] + lhs[14]*rhs[7];
	result[7] = lhs[3]*rhs[4] + lhs[7]*rhs[5] + lhs[11]*rhs[6] + lhs[15]*rhs[7];

	result[8] = lhs[0]*rhs[8] + lhs[4]*rhs[9] + lhs[8]*rhs[10] + lhs[12]*rhs[11];
	result[9] = lhs[1]*rhs[8] + lhs[5]*rhs[9] + lhs[9]*rhs[10] + lhs[13]*rhs[11];
	result[10] = lhs[2]*rhs[8] + lhs[6]*rhs[9] + lhs[10]*rhs[10] + lhs[14]*rhs[11];
	result[11] = lhs[3]*rhs[8] + lhs[7]*rhs[9] + lhs[11]*rhs[10] + lhs[15]*rhs[11];

	result[12] = lhs[0]*rhs[12] + lhs[4]*rhs[13] + lhs[8]*rhs[14] + lhs[12]*rhs[15];
	result[13] = lhs[1]*rhs[12] + lhs[5]*rhs[13] + lhs[9]*rhs[14] + lhs[13]*rhs[15];
	result[14] = lhs[2]*rhs[12] + lhs[6]*rhs[13] + lhs[10]*rhs[14] + lhs[14]*rhs[15];
	result[15] = lhs[3]*rhs[12] + lhs[7]*rhs[13] + lhs[11]*rhs[14] + lhs[15]*rhs[15];

	return result;
}





/**************************************************************/
/*
   Simple Helper Functions
*/
/**************************************************************/




VkShaderModule VkCoffeeLoadShader(const char *fileName, 
                                  VkDevice device)
{
	// did you run buildshaders.bat???
	// can't find the file ;)
	FILE *fp = fopen(fileName, "rb");
	DBG_ASSERT(fp);

	fseek(fp, 0L, SEEK_END);
	size_t size = ftell(fp);
	fseek(fp, 0L, SEEK_SET);

	//shaderCode = malloc(size);
	char *shaderCode = new char[size];
	size_t retval = fread(shaderCode, size, 1, fp);
	DBG_ASSERT(retval == 1);
	DBG_ASSERT(size > 0);

	fclose(fp);

	
	VkShaderModuleCreateInfo moduleCreateInfo;
	moduleCreateInfo.sType		= VK_STRUCTURE_TYPE_SHADER_MODULE_CREATE_INFO;
	moduleCreateInfo.pNext		= NULL;
	moduleCreateInfo.codeSize	= size;
	moduleCreateInfo.pCode		= (uint32_t*)shaderCode;
	moduleCreateInfo.flags		= 0;

	VkShaderModule shaderModule;
	VK_CHECK_RESULT(vkCreateShaderModule(device, &moduleCreateInfo, NULL, &shaderModule));

	delete[] shaderCode;
	shaderCode = NULL;

	return shaderModule;
}// End VkCoffeeLoadShader(..)






/**************************************************************/
/*
   Helper class to wrap the shared Vulkan components
   
   i.e., instance, VkPhysicalDevice, queue, which are shared

   VkCoffeeDevice
*/
/**************************************************************/




class VkCoffeeDevice
{
public:
   	VkInstance       instance			   = NULL;
	VkPhysicalDevice physicalDevice        = VK_NULL_HANDLE;					     
	VkDevice         device                = NULL;
	VkQueue			 queue				   = NULL;
	const char*		 name				   = "Vulkan Demo - Coffee Break Series Extension";

	VkCoffeeDevice()
	{
		VkResult err;

		// Vulkan instance
		VkApplicationInfo appInfo					= {};
		appInfo.sType								= VK_STRUCTURE_TYPE_APPLICATION_INFO;
		appInfo.pApplicationName					= name;
		appInfo.pEngineName							= name;
		appInfo.apiVersion							= VK_API_VERSION_1_0;

		std::vector<const char*> enabledExtensions	= { VK_KHR_SURFACE_EXTENSION_NAME };

		// Enable surface extensions depending on os
		enabledExtensions.push_back(VK_KHR_WIN32_SURFACE_EXTENSION_NAME);

		VkInstanceCreateInfo instanceCreateInfo		= {};
		instanceCreateInfo.sType					= VK_STRUCTURE_TYPE_INSTANCE_CREATE_INFO;
		instanceCreateInfo.pNext					= NULL;
		instanceCreateInfo.pApplicationInfo			= &appInfo;
		instanceCreateInfo.enabledExtensionCount	= (uint32_t)enabledExtensions.size();
		instanceCreateInfo.ppEnabledExtensionNames	= enabledExtensions.data();
		err = vkCreateInstance(&instanceCreateInfo, nullptr, &instance);


		// Physical device
		uint32_t gpuCount = 0;
		// Get number of available physical devices
		err = 
		vkEnumeratePhysicalDevices(instance, &gpuCount, nullptr);
		DBG_ASSERT(err==VK_SUCCESS);
		DBG_ASSERT(gpuCount > 0);
		// Enumerate devices

		// Chapter 5
		// Query how many devices are present in the system
		uint32_t deviceCount = 0;
		err = 
		vkEnumeratePhysicalDevices(instance, &deviceCount, NULL);
		DBG_ASSERT_VULKAN_MSG(err, 
		  "Failed to query the number of physical devices present");

		// There has to be at least one device present
		DBG_ASSERT_MSG(0 != deviceCount, 
		  "Couldn't detect any device present with Vulkan support");

		// Get the physical devices
		vector<VkPhysicalDevice> physicalDevices(deviceCount);
		err = 
		vkEnumeratePhysicalDevices(instance, &deviceCount, &physicalDevices[0]);
		DBG_ASSERT_VULKAN_MSG(err, 
		  "Faied to enumerate physical devices present");
		DBG_ASSERT(physicalDevices.size()>0);
	
		// Use the first available device
		physicalDevice = physicalDevices[0];

		#if 0 // debug
		// Enumerate all physical devices and print out the details
		for (uint32_t i = 0; i < deviceCount; ++i) 
		{
		  VkPhysicalDeviceProperties deviceProperties;
		  memset(&deviceProperties, 0, sizeof deviceProperties);
		  vkGetPhysicalDeviceProperties(physicalDevices[i], &deviceProperties);
		  dprintf("Driver Version: %d\n", deviceProperties.driverVersion);
		  dprintf("Device Name:    %s\n", deviceProperties.deviceName);
		  dprintf("Device Type:    %d\n", deviceProperties.deviceType);
		  dprintf("API Version:    %d.%d.%d\n",
			(deviceProperties.apiVersion>>22)&0x3FF,
			(deviceProperties.apiVersion>>12)&0x3FF,
			(deviceProperties.apiVersion&0xFFF));
		}//End for i
		#endif

		// Chapter 6
		// Fill up the physical device memory properties: 
		VkPhysicalDeviceMemoryProperties memoryProperties;
		vkGetPhysicalDeviceMemoryProperties( physicalDevice, &memoryProperties );

		// Here's where we initialize our queues
		VkDeviceQueueCreateInfo queueCreateInfo = {};
		queueCreateInfo.sType = VK_STRUCTURE_TYPE_DEVICE_QUEUE_CREATE_INFO;
		// Use the first queue family in the family list
		queueCreateInfo.queueFamilyIndex = 0;
		queueCreateInfo.queueCount       = 1;
		float queuePriorities[]          = { 1.0f };
		queueCreateInfo.pQueuePriorities = queuePriorities;

		const char *deviceExtensions[] = { "VK_KHR_swapchain" };
		const char *layers[]           = { "VK_LAYER_NV_optimus" };

		VkDeviceCreateInfo dci = {};
		dci.sType = VK_STRUCTURE_TYPE_DEVICE_CREATE_INFO;
		// Set queue(s) into the device
		dci.queueCreateInfoCount    = 1;
		dci.pQueueCreateInfos       = &queueCreateInfo;
		dci.enabledLayerCount       = 0;
		dci.ppEnabledLayerNames     = layers;
		dci.enabledExtensionCount   = 1;
		dci.ppEnabledExtensionNames = deviceExtensions;

		VkPhysicalDeviceFeatures features = {};
		features.shaderClipDistance = VK_TRUE;
		dci.pEnabledFeatures        = &features;

		// Ideally, we'd want to enumerate and find the best
		// device, however, we just use the first device 
		// `physicalDevices[0]' for our sample, which we
		// stored in the previous chapter
		err = 
		vkCreateDevice( physicalDevice, 
		                &dci, 
						NULL, 
						&device );
		DBG_ASSERT_VULKAN_MSG( err, 
			"Failed to create logical device!" );


		// Get a graphics queue from the device
		vkGetDeviceQueue(device, 
		                 0, 
						 0, 
						 &queue);

	}

	uint32_t GetMemoryType(uint32_t typeBits, 
	                       VkMemoryPropertyFlags properties, 
						   VkBool32 *memTypeFound = nullptr)
	{
		// read the device memory properties
		VkPhysicalDeviceMemoryProperties memoryProperties;
		vkGetPhysicalDeviceMemoryProperties( physicalDevice, &memoryProperties );

		for (uint32_t i = 0; i < memoryProperties.memoryTypeCount; i++)
		{
			if ((typeBits & 1) == 1)
			{
				if ((memoryProperties.memoryTypes[i].propertyFlags & properties) == properties)
				{
					if (memTypeFound)
					{
						*memTypeFound = true;
					}
					return i;
				}
			}
			typeBits >>= 1;
		}

		if (memTypeFound)
		{
			*memTypeFound = false;
			return 0;
		}
		else
		{
			__debugbreak();
			throw std::runtime_error("Could not find a matching memory type");
		}
	}

	
	VkResult CreateBuffer(VkBufferUsageFlags      usageFlags, 
		                    VkMemoryPropertyFlags memoryPropertyFlags, 
							VkDeviceSize          size, 
							VkBuffer *            buffer, 
							VkDeviceMemory *      memory, 
							void *                data = nullptr)
	{
		// Create the buffer handle
		VkBufferCreateInfo bufferCreateInfo = {};
		bufferCreateInfo.sType = VK_STRUCTURE_TYPE_BUFFER_CREATE_INFO;
		bufferCreateInfo.pNext = NULL;
		bufferCreateInfo.usage = usageFlags;
		bufferCreateInfo.size  = size;
		bufferCreateInfo.flags = 0;

		//VkBufferCreateInfo bufferCreateInfo = vkTools::initializers::bufferCreateInfo(usageFlags, size);
		bufferCreateInfo.sharingMode = VK_SHARING_MODE_EXCLUSIVE;
		vkCreateBuffer(device, &bufferCreateInfo, nullptr, buffer);

		// Create the memory backing up the buffer handle
		VkMemoryRequirements memReqs;
		VkMemoryAllocateInfo memAlloc = {};
		memAlloc.sType           = VK_STRUCTURE_TYPE_MEMORY_ALLOCATE_INFO;
		memAlloc.pNext           = NULL;
		memAlloc.allocationSize  = 0;
		memAlloc.memoryTypeIndex = 0;

		//VkMemoryAllocateInfo memAlloc = vkTools::initializers::memoryAllocateInfo();
		vkGetBufferMemoryRequirements(device, *buffer, &memReqs);
		memAlloc.allocationSize = memReqs.size;
		// Find a memory type index that fits the properties of the buffer
		memAlloc.memoryTypeIndex = GetMemoryType(memReqs.memoryTypeBits, memoryPropertyFlags);
		vkAllocateMemory(device, &memAlloc, nullptr, memory);
			
		// If a pointer to the buffer data has been passed, map the buffer and copy over the data
		if (data != nullptr)
		{
			void *mapped;
			vkMapMemory(device, *memory, 0, size, 0, &mapped);
			memcpy(mapped, data, (size_t)size);
			vkUnmapMemory(device, *memory);
		}

		// Attach the memory to the buffer object
		vkBindBufferMemory(device, *buffer, *memory, 0);

		return VK_SUCCESS;
	}
};






/**************************************************************/
/*
   Create test meshes (plane,quad,cube) - for debugging/testing

   VkCoffeeMesh
*/
/**************************************************************/


class VkCoffeeMesh
{
public:
	struct Vertex 
	{
		float pos[3];
		float uv[2];
		float col[3];
		float normal[3];
	};

	struct VerticeDescription
	{
		VkPipelineVertexInputStateCreateInfo           inputState;
		std::vector<VkVertexInputBindingDescription>   bindingDescriptions;
		std::vector<VkVertexInputAttributeDescription> attributeDescriptions;
	};

	VkBuffer			vbuf  = VK_NULL_HANDLE;
	VkDeviceMemory		vmem  = VK_NULL_HANDLE;
	size_t				vsize = 0;
	
	VkBuffer			ibuf  = VK_NULL_HANDLE;
	VkDeviceMemory		imem  = VK_NULL_HANDLE;
	size_t				isize = 0;

	uint32_t			indexCount;
	Vector3				dim;
	VerticeDescription  verticeDescription;



	enum MeshType
	{
		Cube,
		Quad,
		Plane
	};

	VkCoffeeMesh( VkCoffeeDevice*	coffeeDevice,
				  MeshType			mt )
	{
		switch (mt)
		{
			case MeshType::Cube:
			{
				CreateCube(coffeeDevice);
			}
			break;

			case MeshType::Plane:
			{
				CreatePlane(coffeeDevice);
			}
			break;

			case MeshType::Quad:
			{
				CreateQuad(coffeeDevice);
			}
			break;

			default:
			{
				__debugbreak();
			}
			break;
		}// End switch(..)
		
		SetupVertexDescriptions();

	}// VkCoffeeMesh(..)


	void SetupVertexDescriptions()
	{
		// Binding description
		verticeDescription.bindingDescriptions.resize(1);

		VkVertexInputBindingDescription vInputBindDescription = {};
		vInputBindDescription.binding   = VERTEX_BUFFER_BIND_ID;
		vInputBindDescription.stride    = sizeof(Vertex);
		vInputBindDescription.inputRate = VK_VERTEX_INPUT_RATE_VERTEX;
		verticeDescription.bindingDescriptions[0] = vInputBindDescription;

		// Attribute descriptions
		verticeDescription.attributeDescriptions.resize(4);

		// Location 0 : Position
		verticeDescription.attributeDescriptions[0].location = 0;
		verticeDescription.attributeDescriptions[0].binding  = VERTEX_BUFFER_BIND_ID;
		verticeDescription.attributeDescriptions[0].format   = VK_FORMAT_R32G32B32_SFLOAT;
		verticeDescription.attributeDescriptions[0].offset   = 0;

		// Location 1 : Texture coordinates
		verticeDescription.attributeDescriptions[1].location = 1;
		verticeDescription.attributeDescriptions[1].binding  = VERTEX_BUFFER_BIND_ID;
		verticeDescription.attributeDescriptions[1].format   = VK_FORMAT_R32G32_SFLOAT;
		verticeDescription.attributeDescriptions[1].offset   = sizeof(float) * 3;

		// Location 2 : Color
		verticeDescription.attributeDescriptions[2].location = 2;
		verticeDescription.attributeDescriptions[2].binding  = VERTEX_BUFFER_BIND_ID;
		verticeDescription.attributeDescriptions[2].format   = VK_FORMAT_R32G32B32_SFLOAT;
		verticeDescription.attributeDescriptions[2].offset   = sizeof(float) * 5;

		// Location 3 : Normal
		verticeDescription.attributeDescriptions[3].location = 3;
		verticeDescription.attributeDescriptions[3].binding  = VERTEX_BUFFER_BIND_ID;
		verticeDescription.attributeDescriptions[3].format   = VK_FORMAT_R32G32B32_SFLOAT;
		verticeDescription.attributeDescriptions[3].offset   = sizeof(float) * 8;

		VkPipelineVertexInputStateCreateInfo pipelineVertexInputStateCreateInfo = {};
		pipelineVertexInputStateCreateInfo.sType = VK_STRUCTURE_TYPE_PIPELINE_VERTEX_INPUT_STATE_CREATE_INFO;
		pipelineVertexInputStateCreateInfo.pNext = NULL;

		verticeDescription.inputState                                 = pipelineVertexInputStateCreateInfo;
		verticeDescription.inputState.vertexBindingDescriptionCount   = verticeDescription.bindingDescriptions.size();
		verticeDescription.inputState.pVertexBindingDescriptions      = verticeDescription.bindingDescriptions.data();
		verticeDescription.inputState.vertexAttributeDescriptionCount = verticeDescription.attributeDescriptions.size();
		verticeDescription.inputState.pVertexAttributeDescriptions    = verticeDescription.attributeDescriptions.data();
	}// End SetupVertexDescriptions(..)


	void CreateQuad(VkCoffeeDevice* coffeeDevice)
	{		
	    // Setup vertices for a single uv-mapped quad
#define QUAD_COLOUR_NORMAL { 1.0f, 1.0f, 1.0f }, { 0.0f, 0.0f, 1.0f }
		std::vector<Vertex> vertexBuffer =
		{
			{ { 1.0f, 1.0f, 0.0f },{ 1.0f, 1.0f }, QUAD_COLOUR_NORMAL },
			{ { 0.0f, 1.0f, 0.0f },{ 0.0f, 1.0f }, QUAD_COLOUR_NORMAL },
			{ { 0.0f, 0.0f, 0.0f },{ 0.0f, 0.0f }, QUAD_COLOUR_NORMAL },
			{ { 1.0f, 0.0f, 0.0f },{ 1.0f, 0.0f }, QUAD_COLOUR_NORMAL }
		};
#undef QUAD_COLOUR_NORMAL

		CreateBuffer(
			coffeeDevice,
			VK_BUFFER_USAGE_VERTEX_BUFFER_BIT,
			VK_MEMORY_PROPERTY_HOST_VISIBLE_BIT,
			vertexBuffer.size() * sizeof(Vertex),
			vertexBuffer.data(),
			&vbuf,
			&vmem);

		// Setup indices
		std::vector<uint32_t> indexBuffer = { 0,1,2, 2,3,0 };
		indexCount = indexBuffer.size();

		CreateBuffer(
			coffeeDevice,
			VK_BUFFER_USAGE_INDEX_BUFFER_BIT,
			VK_MEMORY_PROPERTY_HOST_VISIBLE_BIT,
			indexBuffer.size() * sizeof(uint32_t),
			indexBuffer.data(),
			&ibuf,
			&imem);
	}


	void CreatePlane(VkCoffeeDevice* coffeeDevice)
	{		
	    // Setup vertices for a single uv-mapped quad
		float s = 3.0f;
		const float h = -1.0f;
#define PLANE_COLOUR_NORMAL { 1.0f, 1.0f, 1.0f }, { 0.0f, 1.0f, 0.0f }
		std::vector<Vertex> vertexBuffer =
		{

			{ {  s, h,  s }, { 1.0f, 1.0f }, PLANE_COLOUR_NORMAL },
			{ { -s, h,  s }, { 0.0f, 1.0f }, PLANE_COLOUR_NORMAL },
			{ { -s, h, -s }, { 0.0f, 0.0f }, PLANE_COLOUR_NORMAL },
			{ {  s, h, -s }, { 1.0f, 0.0f }, PLANE_COLOUR_NORMAL }
		};
#undef PLANE_COLOUR_NORMAL

		CreateBuffer(
			coffeeDevice,
			VK_BUFFER_USAGE_VERTEX_BUFFER_BIT,
			VK_MEMORY_PROPERTY_HOST_VISIBLE_BIT,
			vertexBuffer.size() * sizeof(Vertex),
			vertexBuffer.data(),
			&vbuf,
			&vmem);

		// Setup indices
		std::vector<uint32_t> indexBuffer = { 0,1,2, 2,3,0 };
		indexCount = indexBuffer.size();

		CreateBuffer(
			coffeeDevice,
			VK_BUFFER_USAGE_INDEX_BUFFER_BIT,
			VK_MEMORY_PROPERTY_HOST_VISIBLE_BIT,
			indexBuffer.size() * sizeof(uint32_t),
			indexBuffer.data(),
			&ibuf,
			&imem);
	}

	void CreateCube(VkCoffeeDevice* coffeeDevice)
	{
		std::vector<Vertex> vertexBuffer =
		{
			{ {	-1.0f, -1.0f, -1.0f }, { 0.0f, 1.0f }, { 1.0f, 0.0f, 0.0f}, { 0.0, 0.0, 1.0 } },  
			{ {  1.0f, -1.0f, -1.0f }, { 1.0f, 1.0f }, { 1.0f, 0.0f, 0.0f}, { 0.0, 0.0, 1.0 } },  
			{ { -1.0f,  1.0f, -1.0f }, { 0.0f, 0.0f }, { 1.0f, 0.0f, 0.0f}, { 0.0, 0.0, 1.0 } },  
			{ {  1.0f,  1.0f, -1.0f }, { 1.0f, 0.0f }, { 1.0f, 0.0f, 0.0f}, { 0.0, 0.0, 1.0 } },  
																							 
			{ { -1.0f, -1.0f,  1.0f }, { 0.0f, 1.0f }, { 0.0f, 1.0f, 0.0f}, {  1.0, 0.0, 0.0 } },  
			{ { -1.0f, -1.0f, -1.0f }, { 1.0f, 1.0f }, { 0.0f, 1.0f, 0.0f}, {  1.0, 0.0, 0.0 } }, 
			{ { -1.0f,  1.0f,  1.0f }, { 0.0f, 0.0f }, { 0.0f, 1.0f, 0.0f}, {  1.0, 0.0, 0.0 } }, 
			{ { -1.0f,  1.0f, -1.0f }, { 1.0f, 0.0f }, { 0.0f, 1.0f, 0.0f}, {  1.0, 0.0, 0.0 } }, 
																							
			{ {  1.0f, -1.0f,  1.0f }, { 0.0f, 1.0f }, { 0.0f, 0.0f, 1.0f}, { 0.0, 0.0, -1.0 } }, 
			{ { -1.0f, -1.0f,  1.0f }, { 1.0f, 1.0f }, { 0.0f, 0.0f, 1.0f}, { 0.0, 0.0, -1.0 } }, 
			{ {  1.0f,  1.0f,  1.0f }, { 0.0f, 0.0f }, { 0.0f, 0.0f, 1.0f}, { 0.0, 0.0, -1.0 } }, 
			{ { -1.0f,  1.0f,  1.0f }, { 1.0f, 0.0f }, { 0.0f, 0.0f, 1.0f}, { 0.0, 0.0, -1.0 } }, 
																							 
			{ {  1.0f, -1.0f, -1.0f }, { 0.0f, 1.0f }, { 1.0f, 0.0f, 1.0f}, {0.0, -1.0, 0.0 } }, 
			{ {  1.0f, -1.0f,  1.0f }, { 1.0f, 1.0f }, { 1.0f, 0.0f, 1.0f}, {0.0, -1.0, 0.0 } }, 
			{ {  1.0f,  1.0f, -1.0f }, { 0.0f, 0.0f }, { 1.0f, 0.0f, 1.0f}, {0.0, -1.0, 0.0 } }, 
			{ {  1.0f,  1.0f,  1.0f }, { 1.0f, 0.0f }, { 1.0f, 0.0f, 1.0f}, {0.0, -1.0, 0.0 } }, 
																							 
			{ { -1.0f,  1.0f, -1.0f }, { 0.0f, 1.0f }, { 1.0f, 1.0f, 0.0f}, {-1.0, 0.0, 0.0 } }, 
			{ {  1.0f,  1.0f, -1.0f }, { 1.0f, 1.0f }, { 1.0f, 1.0f, 0.0f}, {-1.0, 0.0, 0.0 } }, 
			{ { -1.0f,  1.0f,  1.0f }, { 0.0f, 0.0f }, { 1.0f, 1.0f, 0.0f}, {-1.0, 0.0, 0.0 } }, 
			{ {  1.0f,  1.0f,  1.0f }, { 1.0f, 0.0f }, { 1.0f, 1.0f, 0.0f}, {-1.0, 0.0, 0.0 } }, 
																							 
			{ { -1.0f, -1.0f,  1.0f }, { 0.0f, 1.0f }, { 1.0f, 1.0f, 1.0f}, { 0.0, -1.0, 0.0 } }, 
			{ {  1.0f, -1.0f,  1.0f }, { 1.0f, 1.0f }, { 1.0f, 1.0f, 1.0f}, { 0.0, -1.0, 0.0 } }, 
			{ { -1.0f, -1.0f, -1.0f }, { 0.0f, 0.0f }, { 1.0f, 1.0f, 1.0f}, { 0.0, -1.0, 0.0 } }, 
			{ {  1.0f, -1.0f, -1.0f }, { 1.0f, 0.0f }, { 1.0f, 1.0f, 1.0f}, { 0.0, -1.0, 0.0 } } 
		};

		std::vector<uint32_t> indexBuffer = 
		{   0, 1, 3,  0, 3, 2,         //Face front
			4, 5, 7,  4, 7, 6,         //Face right
			8, 9, 11, 8, 11,10,        //...
			12,13,15, 12,15,14,
			16,17,19, 16,19,18,
			20,21,23, 20,23,22 
		};

		CreateBuffer(
			coffeeDevice,
			VK_BUFFER_USAGE_VERTEX_BUFFER_BIT,
			VK_MEMORY_PROPERTY_HOST_VISIBLE_BIT,
			vertexBuffer.size() * sizeof(Vertex),
			vertexBuffer.data(),
			&vbuf,
			&vmem);

		// Setup indices
		indexCount = indexBuffer.size();

		CreateBuffer(
			coffeeDevice,
			VK_BUFFER_USAGE_INDEX_BUFFER_BIT,
			VK_MEMORY_PROPERTY_HOST_VISIBLE_BIT,
			indexBuffer.size() * sizeof(uint32_t),
			indexBuffer.data(),
			&ibuf,
			&imem);

	}

	VkBool32 CreateBuffer(VkCoffeeDevice*		coffeeDevice,
	                      VkBufferUsageFlags    usageFlags, 
	                      VkMemoryPropertyFlags memoryPropertyFlags, 
						  VkDeviceSize          size, 
						  void *                data, 
						  VkBuffer *            buffer, 
						  VkDeviceMemory *      memory)
	{
		VkMemoryRequirements memReqs;
		VkMemoryAllocateInfo memAlloc = {};
		memAlloc.sType           = VK_STRUCTURE_TYPE_MEMORY_ALLOCATE_INFO;
		memAlloc.pNext           = NULL;
		memAlloc.allocationSize  = 0;
		memAlloc.memoryTypeIndex = 0;

		VkBufferCreateInfo bufferCreateInfo = {};
		bufferCreateInfo.sType = VK_STRUCTURE_TYPE_BUFFER_CREATE_INFO;
		bufferCreateInfo.pNext = NULL;
		bufferCreateInfo.usage = usageFlags;
		bufferCreateInfo.size  = size;
		bufferCreateInfo.flags = 0;


		VK_CHECK_RESULT(vkCreateBuffer(coffeeDevice->device, &bufferCreateInfo, nullptr, buffer));

		vkGetBufferMemoryRequirements(coffeeDevice->device, *buffer, &memReqs);
		memAlloc.allocationSize  = memReqs.size;
		memAlloc.memoryTypeIndex = coffeeDevice->GetMemoryType(memReqs.memoryTypeBits, memoryPropertyFlags);
		VK_CHECK_RESULT(vkAllocateMemory(coffeeDevice->device, &memAlloc, nullptr, memory));
		if (data != nullptr)
		{
			void *mapped;
			VK_CHECK_RESULT(vkMapMemory(coffeeDevice->device, *memory, 0, size, 0, &mapped));
			memcpy(mapped, data, (size_t)size);
			vkUnmapMemory(coffeeDevice->device, *memory);
		}
		VK_CHECK_RESULT(vkBindBufferMemory(coffeeDevice->device, *buffer, *memory, 0));

		return true;
	}

};








/**************************************************************/
/*
   Create/setup and wrap all the swap-chain and surface
   variables/calls together in one place

   VkCoffeeSwapChain
*/
/**************************************************************/


class VkCoffeeSwapChain
{
public:
	VkSwapchainKHR   swapChain             = NULL;
	VkSurfaceKHR     surface               = NULL;
	int				 width				   = 0;
	int				 height				   = 0;
	uint32_t		 imageCount			   = 0;
	VkImage*		 images;
	VkImageView*     imageViews;

	// Index of the deteced graphics and presenting device queue
	uint32_t queueNodeIndex = UINT32_MAX;

	VkCoffeeSwapChain(VkInstance	   instance,
	                  VkDevice		   device,
					  VkPhysicalDevice physicalDevice,
					  HWND			   windowHandle )
	{
		// setup parameters for our new windows 
		// surface we'll render into:
		VkWin32SurfaceCreateInfoKHR sci = {};
		sci.sType	  = VK_STRUCTURE_TYPE_WIN32_SURFACE_CREATE_INFO_KHR;
		// parameter is NULL, GetModuleHandle returns a handle 
		// to the file used to create the calling process 
		sci.hinstance = GetModuleHandle(NULL);
		// We should use our `created' window handle (HWND)
		sci.hwnd	  = windowHandle;

		VkResult result = 
		vkCreateWin32SurfaceKHR( instance, &sci, NULL, &surface );

		DBG_ASSERT_VULKAN_MSG( result, 
		  "Could not create surface." );
		DBG_ASSERT(surface!=NULL);


		// swap-chain creation:
		VkSurfaceCapabilitiesKHR surfaceCapabilities = {};
		vkGetPhysicalDeviceSurfaceCapabilitiesKHR( physicalDevice, 
		                                           surface, 
												   &surfaceCapabilities );

		VkExtent2D surfaceResolution =  surfaceCapabilities.currentExtent;
		width  = surfaceResolution.width;
		height = surfaceResolution.height;

		imageCount = 2; // double buffering

		VkSwapchainCreateInfoKHR ssci = {};
		ssci.sType            = VK_STRUCTURE_TYPE_SWAPCHAIN_CREATE_INFO_KHR;
		ssci.pNext			  = NULL;
		ssci.surface          = surface;
		// Should be 2 for `double' buffering
		ssci.minImageCount    = imageCount;
		ssci.imageFormat      = VK_FORMAT_B8G8R8A8_UNORM;
		ssci.imageColorSpace  = VK_COLORSPACE_SRGB_NONLINEAR_KHR;
		ssci.imageExtent      = surfaceResolution;
		ssci.imageArrayLayers = 1;
		ssci.imageUsage       = VK_IMAGE_USAGE_COLOR_ATTACHMENT_BIT; // VK_IMAGE_USAGE_COLOR_ATTACHMENT_BIT | VK_IMAGE_USAGE_TRANSFER_SRC_BIT;
		ssci.imageSharingMode = VK_SHARING_MODE_EXCLUSIVE;
		ssci.preTransform     = VK_SURFACE_TRANSFORM_IDENTITY_BIT_KHR;
		ssci.compositeAlpha   = VK_COMPOSITE_ALPHA_OPAQUE_BIT_KHR;
		ssci.presentMode      = VK_PRESENT_MODE_MAILBOX_KHR;
		// If we want clipping outside the extents
		ssci.clipped          = true;     
		ssci.oldSwapchain     = NULL;

		result = 
		vkCreateSwapchainKHR( device, 
							  &ssci, 
							  NULL, 
							  &swapChain );
		DBG_ASSERT_VULKAN_MSG( result, 
		  "Failed to create swapchain." );


		// this should be 2 for double-buffering
		images = new VkImage[imageCount];  

		// link the images to the swapchain  
		result =
		vkGetSwapchainImagesKHR( device, 
								 swapChain, 
								 &imageCount, 
								 images );
		DBG_ASSERT_VULKAN_MSG( result, 
			"Failed to create swap-chain images" );

		// Should be 2 for double buffering
		imageViews = new VkImageView[imageCount];
		for( uint32_t i = 0; i < imageCount; ++i ) 
		{
		// create VkImageViews for our swap chain 
		// VkImages buffers:
		VkImageViewCreateInfo ivci = {};
		ivci.sType						    = VK_STRUCTURE_TYPE_IMAGE_VIEW_CREATE_INFO;
		ivci.viewType					    = VK_IMAGE_VIEW_TYPE_2D;
		ivci.format						    = VK_FORMAT_B8G8R8A8_UNORM;
		ivci.components.r					= VK_COMPONENT_SWIZZLE_R;
		ivci.components.g					= VK_COMPONENT_SWIZZLE_G;
		ivci.components.b					= VK_COMPONENT_SWIZZLE_B;
		ivci.components.a					= VK_COMPONENT_SWIZZLE_A;
		ivci.subresourceRange.aspectMask     = VK_IMAGE_ASPECT_COLOR_BIT;
		ivci.subresourceRange.baseMipLevel   = 0;
		ivci.subresourceRange.levelCount     = 1;
		ivci.subresourceRange.baseArrayLayer = 0;
		ivci.subresourceRange.layerCount     = 1;
		ivci.image                           = images[i];

		VkResult result = 
		vkCreateImageView( device, &ivci, NULL, &imageViews[i] );

		DBG_ASSERT_VULKAN_MSG( result, 
		  "Could not create ImageView." );
		}// End for i

	}

	VkResult AcquireNextImage(VkDevice    device,
	                          VkSemaphore presentCompleteSemaphore, 
	                          uint32_t *  imageIndex)
	{
		return
		vkAcquireNextImageKHR( device, 
							   swapChain, 
							   UINT64_MAX,
							   presentCompleteSemaphore, 
							   VK_NULL_HANDLE, 
							   imageIndex );

	}


	VkResult QueuePresent(VkQueue     queue, 
	                      uint32_t    imageIndex, 
						  VkSemaphore waitSemaphore = VK_NULL_HANDLE)
	{
		VkPresentInfoKHR presentInfo = {};
		presentInfo.sType           = VK_STRUCTURE_TYPE_PRESENT_INFO_KHR;
		presentInfo.pNext           = NULL;
		presentInfo.swapchainCount  = 1;
		presentInfo.pSwapchains     = &swapChain;
		presentInfo.pImageIndices   = &imageIndex;
		// Check if a wait semaphore has been specified to wait for before presenting the image
		if (waitSemaphore != VK_NULL_HANDLE)
		{
			presentInfo.pWaitSemaphores = &waitSemaphore;
			presentInfo.waitSemaphoreCount = 1;
		}
		return vkQueuePresentKHR(queue, &presentInfo);
	}

};





/**************************************************************/
/*
   Create/setup and contain the texture information
   Currently uses a simple test pattern, however, can be 
   modified with a few lines of code to load a texture

   VkCoffeeTexture
*/
/**************************************************************/

class VkCoffeeTexture
{
public:

	VkSampler				sampler			    = NULL;
	VkImage					image			    = NULL;
	VkImageView				imageView		    = NULL;
	VkDeviceMemory			imageMemory		    = { 0 };
	VkDescriptorImageInfo	descriptorImageInfo;
	int						width			    = 0;
	int						height			    = 0;

	VkCoffeeTexture(VkDevice		 device,
					VkPhysicalDevice physicalDevice)
	{
		VkSamplerCreateInfo samplerCreateInfo = {};
		samplerCreateInfo.sType             = VK_STRUCTURE_TYPE_SAMPLER_CREATE_INFO;
		samplerCreateInfo.magFilter         = VK_FILTER_LINEAR;
		samplerCreateInfo.minFilter         = VK_FILTER_LINEAR;
		samplerCreateInfo.mipmapMode        = VK_SAMPLER_MIPMAP_MODE_LINEAR;
		samplerCreateInfo.addressModeU      = VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE;
		samplerCreateInfo.addressModeV      = VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE;
		samplerCreateInfo.addressModeW      = VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE;
		samplerCreateInfo.mipLodBias       = 0;
		samplerCreateInfo.anisotropyEnable = VK_FALSE;
		samplerCreateInfo.minLod           = 0;
		samplerCreateInfo.maxLod           = 5;
		samplerCreateInfo.borderColor      = VK_BORDER_COLOR_FLOAT_TRANSPARENT_BLACK;
		samplerCreateInfo.unnormalizedCoordinates = VK_FALSE;

    
		VkResult result = 
		vkCreateSampler( device, 
						 &samplerCreateInfo, 
						 NULL, 
						 &sampler );

		DBG_ASSERT_VULKAN_MSG( result, 
			"Failed to create sampler." );


		width  = 512;
		height = 512;
		void* imageData   = (void *) new DWORD[ width * height ];

	
		// generate a pretty test pattern
		for (int y=0; y<height; ++y)
		{
			for (int x=0; x<width; ++x)
			{
				int cc = (x^y)   &0xFF;
				int r  = (cc>>16)&0xFF;
				int g  = (cc>>8) &0xFF;
				int b  = (cc>>0) &0xFF ;
				// beautiful xor pattern
				DWORD* pixel = ((DWORD *) imageData) + ( x * height  ) + ( y );
				pixel[0] = cc;
			}// End inner loop
		}// End outer loop

		VkImageCreateInfo textureCreateInfo = {};
		textureCreateInfo.sType         = VK_STRUCTURE_TYPE_IMAGE_CREATE_INFO;
		textureCreateInfo.imageType     = VK_IMAGE_TYPE_2D;
		textureCreateInfo.format        = VK_FORMAT_R8G8B8A8_UNORM;;
		textureCreateInfo.extent.width  = width; 
		textureCreateInfo.extent.height = height; 
		textureCreateInfo.extent.depth  = 1; 
		textureCreateInfo.mipLevels     = 1;
		textureCreateInfo.arrayLayers   = 1;
		textureCreateInfo.samples       = VK_SAMPLE_COUNT_1_BIT;
		textureCreateInfo.tiling        = VK_IMAGE_TILING_LINEAR;
		textureCreateInfo.usage         = VK_IMAGE_USAGE_SAMPLED_BIT;
		textureCreateInfo.sharingMode   = VK_SHARING_MODE_EXCLUSIVE;
		textureCreateInfo.initialLayout = VK_IMAGE_LAYOUT_PREINITIALIZED;

		result = 
		vkCreateImage( device, 
					   &textureCreateInfo, 
					   NULL, 
					   &image );

		DBG_ASSERT_VULKAN_MSG( result, 
			"Failed to create texture image." );

		// allocate and bind image memory:
		VkMemoryRequirements textureMemoryRequirements = {};
		vkGetImageMemoryRequirements( device, 
									  image, 
									  &textureMemoryRequirements );

		VkMemoryAllocateInfo textureImageAllocateInfo = {};
		textureImageAllocateInfo.sType = VK_STRUCTURE_TYPE_MEMORY_ALLOCATE_INFO;
		textureImageAllocateInfo.allocationSize = textureMemoryRequirements.size;

		// read the device memory properties
		VkPhysicalDeviceMemoryProperties memoryProperties;
		vkGetPhysicalDeviceMemoryProperties( physicalDevice, &memoryProperties );

		uint32_t textureMemoryTypeBits = textureMemoryRequirements.memoryTypeBits;
		VkMemoryPropertyFlags tDesiredMemoryFlags = VK_MEMORY_PROPERTY_HOST_VISIBLE_BIT;
		for( uint32_t i = 0; i < 32; ++i ) 
		{
			VkMemoryType memoryType = memoryProperties.memoryTypes[i];
			if( textureMemoryTypeBits & 1 ) 
			{
				if( ( memoryType.propertyFlags & tDesiredMemoryFlags ) == tDesiredMemoryFlags ) 
				{
					textureImageAllocateInfo.memoryTypeIndex = i;
					break;
				}
			}
			textureMemoryTypeBits = textureMemoryTypeBits >> 1;
		}

		result = vkAllocateMemory( device, &textureImageAllocateInfo, NULL, &imageMemory );
		DBG_ASSERT_VULKAN_MSG( result, 
			"Failed to allocate device memory." );

		result = vkBindImageMemory( device, image, imageMemory, 0 );
		DBG_ASSERT_VULKAN_MSG( result, 
			"Failed to bind image memory." );

		// copy our image content:
		void *imageMapped;
		result = vkMapMemory( device, imageMemory, 0, VK_WHOLE_SIZE, 0, &imageMapped );
		DBG_ASSERT_VULKAN_MSG( result, 
			"Failed to map uniform buffer memory." );
    
		memcpy( imageMapped, imageData, sizeof(DWORD) * width * height  );

		VkMappedMemoryRange memoryRange = {};
		memoryRange.sType  = VK_STRUCTURE_TYPE_MAPPED_MEMORY_RANGE;
		memoryRange.memory = imageMemory;
		memoryRange.offset = 0;
		memoryRange.size   = VK_WHOLE_SIZE;
		vkFlushMappedMemoryRanges( device, 1, &memoryRange );
    
		vkUnmapMemory( device, imageMemory );

	
		delete[] imageData;
		imageData = NULL;

		// create and map the texture view
		VkImageViewCreateInfo textureImageViewCreateInfo = {};
		textureImageViewCreateInfo.sType    = VK_STRUCTURE_TYPE_IMAGE_VIEW_CREATE_INFO;
		textureImageViewCreateInfo.image    = image;
		textureImageViewCreateInfo.viewType = VK_IMAGE_VIEW_TYPE_2D;
		textureImageViewCreateInfo.format   = VK_FORMAT_R8G8B8A8_UNORM;//VK_FORMAT_R32G32B32_SFLOAT;
		VkComponentMapping j = { VK_COMPONENT_SWIZZLE_R, 
								 VK_COMPONENT_SWIZZLE_G, 
								 VK_COMPONENT_SWIZZLE_B, 
								 VK_COMPONENT_SWIZZLE_A };
		textureImageViewCreateInfo.components                      = j; 
		textureImageViewCreateInfo.subresourceRange.aspectMask     = VK_IMAGE_ASPECT_COLOR_BIT;
		textureImageViewCreateInfo.subresourceRange.baseMipLevel   = 0;
		textureImageViewCreateInfo.subresourceRange.levelCount     = 1;
		textureImageViewCreateInfo.subresourceRange.baseArrayLayer = 0;
		textureImageViewCreateInfo.subresourceRange.layerCount     = 1;

		result = 
		vkCreateImageView( device, 
						   &textureImageViewCreateInfo, 
						   NULL, 
						   &imageView );

		DBG_ASSERT_VULKAN_MSG( result, 
			"Failed to create image view." );

		// Fill descriptor image info that can be used for setting up descriptor sets
		descriptorImageInfo.imageLayout = VK_IMAGE_LAYOUT_GENERAL;
		descriptorImageInfo.imageView   = imageView;
		descriptorImageInfo.sampler     = sampler;
	}

};





VkPipelineShaderStageCreateInfo LoadShader(std::string fileName, VkShaderStageFlagBits stage, VkShaderModule* shaderModule)
{
	extern VkCoffeeDevice * coffeeDevice;

	VkPipelineShaderStageCreateInfo shaderStage = {};
	shaderStage.sType  = VK_STRUCTURE_TYPE_PIPELINE_SHADER_STAGE_CREATE_INFO;
	shaderStage.stage  = stage;

	shaderStage.module = VkCoffeeLoadShader(fileName.c_str(), coffeeDevice->device);

	shaderStage.pName = "main";
	DBG_ASSERT(shaderStage.module != NULL);
	*shaderModule = shaderStage.module;
	return shaderStage;
}




inline
VkBool32 CreateBuffer(VkBufferUsageFlags    usageFlags, 
	                    VkMemoryPropertyFlags memoryPropertyFlags, 
						VkDeviceSize          size, 
						void *                data, 
						VkBuffer *            buffer, 
						VkDeviceMemory *      memory)
{
	extern VkCoffeeDevice *				coffeeDevice;


	VkMemoryRequirements memReqs;

	VkMemoryAllocateInfo memAlloc = {};
	memAlloc.sType = VK_STRUCTURE_TYPE_MEMORY_ALLOCATE_INFO;
	memAlloc.pNext = NULL;
	memAlloc.allocationSize = 0;
	memAlloc.memoryTypeIndex = 0;

	VkBufferCreateInfo bufferCreateInfo = {};
	bufferCreateInfo.sType = VK_STRUCTURE_TYPE_BUFFER_CREATE_INFO;
	bufferCreateInfo.pNext = NULL;
	bufferCreateInfo.usage = usageFlags;
	bufferCreateInfo.size  = size;
	bufferCreateInfo.flags = 0;


	VK_CHECK_RESULT(vkCreateBuffer(coffeeDevice->device, &bufferCreateInfo, nullptr, buffer));

	vkGetBufferMemoryRequirements(coffeeDevice->device, *buffer, &memReqs);
	memAlloc.allocationSize = memReqs.size;
	memAlloc.memoryTypeIndex = coffeeDevice->GetMemoryType(memReqs.memoryTypeBits, memoryPropertyFlags);
	VK_CHECK_RESULT(vkAllocateMemory(coffeeDevice->device, &memAlloc, nullptr, memory));
	if (data != nullptr)
	{
		void *mapped;
		VK_CHECK_RESULT(vkMapMemory(coffeeDevice->device, *memory, 0, size, 0, &mapped));
		memcpy(mapped, data, (size_t)size);
		vkUnmapMemory(coffeeDevice->device, *memory);
	}
	VK_CHECK_RESULT(vkBindBufferMemory(coffeeDevice->device, *buffer, *memory, 0));

	return true;
}


inline
VkBool32 CreateBuffer(VkBufferUsageFlags      usage, 
	                    VkMemoryPropertyFlags    memoryPropertyFlags, 
						VkDeviceSize             size, 
						void *                   data, 
						VkBuffer *               buffer, 
						VkDeviceMemory *         memory, 
						VkDescriptorBufferInfo * descriptor)
{
	VkBool32 res = CreateBuffer(usage, memoryPropertyFlags, size, data, buffer, memory);
	if (res)
	{
		descriptor->offset = 0;
		descriptor->buffer = *buffer;
		descriptor->range = size;
		return true;
	}
	else
	{
		return false;
	}
}



